/*************************************************************************
 *
 * FILE  main.cpp
 * Copyright (c) 2005, 2006, 2007, 2008 Daniel Kahlin <daniel@kahlin.net>
 * Written by Daniel Kahlin <daniel@kahlin.net>
 * $Id: main.cpp,v 1.97 2008-02-18 21:40:40 tlr Exp $
 *
 * DESCRIPTION 
 *
 ******/

#include <exception>
#include <iostream>
#include <string>
#include <stdio.h>
#include "util/types.hpp"
#include "util/exceptions.hpp"
#include "util/getopt.hpp"
#include "global.hpp"
#include "parport_transfer.hpp"
#include "membuffer.hpp"
// arch dependant
#include "timer.hpp"

#define PROGRAM "dtvtrans"

#ifndef IS_LINUX
#define PORT_DEFAULT "0x378"
#else
#define PORT_DEFAULT "/dev/parport0"
#endif

typedef int (*cmd_func_t)(Getopt&);
int cmd_boot(Getopt& opts);
int cmd_read_mem(Getopt& opts);
int cmd_verify_mem(Getopt& opts);
int cmd_write_mem(Getopt& opts);
int cmd_execute_mem(Getopt& opts);
int cmd_reset(Getopt& opts);
int cmd_clear_port(Getopt& opts);
int cmd_sleep(Getopt& opts);
int cmd_info(Getopt& opts);
int cmd_init(Getopt& opts);
int cmd_load(Getopt& opts);
int cmd_save(Getopt& opts);
int cmd_run(Getopt& opts);
int cmd_exit(Getopt& opts);

void perform_init(Transfer::init_mode_t mode);

// Petscii definitions
#define CBM_CR "\x0d"
#define CBM_CURSOR "\x12 \x92"
#define CBM_NOCURSOR "\x9d \x9d"
#define CBM_READY CBM_CR "READY." CBM_CR

/*************************************************************************
 *
 * NAME  validate_opt
 *
 * DESCRIPTION
 *   Check status of Getopt::Opt and throw if error.
 *
 ******/
void validate_opt(Getopt::Opt& opt)
{
    char tmpstr[256];
    switch(opt.get_status()) {
    case Getopt::Opt::OK:
	break;
    case Getopt::Opt::MISSING:
	snprintf(tmpstr, sizeof(tmpstr), "missing parameter for option -%c",opt.get_opt());
	throw Exception(tmpstr);
    case Getopt::Opt::INVALID:

	snprintf(tmpstr, sizeof(tmpstr), "invalid option -%c",opt.get_opt());
	throw Exception(tmpstr);
    default:
	snprintf(tmpstr, sizeof(tmpstr), "error while parsing option -%c",opt.get_opt());
	throw Exception(tmpstr);
    }
}

/*************************************************************************
 *
 * NAME  get_arg_as_val
 *
 * DESCRIPTION
 *   Get argument as value, and throw if error.
 *
 ******/
uint32_t get_arg_as_val(Getopt::Opt& opt)
{
    const char* str=opt.get_arg();
    char* endptr;
    uint32_t val=std::strtoul(str,&endptr,0);
    if (endptr==str) {
	char tmpstr[256];
	snprintf(tmpstr, sizeof(tmpstr), "invalid number (%s)",str);
	throw Exception(tmpstr);
    }
    return val;
}


/*************************************************************************
 *
 * NAME  TransferFactory
 *
 * DESCRIPTION
 *   Singleton factory object for the Transfer class.
 *
 ******/
Transfer& get_transfer()
{
    if (debug_g)
	printf("get_transfer() called\n");
    static class TransferFactory 
    {
    public:
	TransferFactory() : tf_(0) { }
	~TransferFactory()
	{
	    delete tf_;
	    if (debug_g)
		printf("destroyed Transfer object\n");
	}

	Transfer& get_transfer()
	{
	    if (!tf_) {
		const char* obj="<none>";
		if (true) {
		    tf_ = new ParportTransfer(portbase_g, blocksize_g, safe_setup_dtv_g);
		    obj = "ParportTransfer";
		} else {
		    //tf_ = new Dtv2serTransfer(portbase_g, blocksize_g, safe_setup_dtv_g);
		    //obj = "Dtv2serTransfer";
		}
		if (debug_g) 
		    printf("created Transfer object (%s)\n", obj );
	    }
	    return *tf_;
	}
    private:
	Transfer* tf_;
    } factory;

    return factory.get_transfer();
}

/*************************************************************************
 *
 * NAME  main
 *
 * DESCRIPTION
 *   main entry point.
 *
 ******/
int main(int argc, char *argv[])
{
    int ret=1;

    std::map<std::string,cmd_func_t> cmds;
    cmds.insert(std::make_pair<std::string,cmd_func_t>("boot",cmd_boot));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("bt",cmd_boot));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("b",cmd_boot));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("read",cmd_read_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("rd",cmd_read_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("r",cmd_read_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("write",cmd_write_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("wr",cmd_write_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("w",cmd_write_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("verify",cmd_verify_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("vfy",cmd_verify_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("sys",cmd_execute_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("go",cmd_execute_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("g",cmd_execute_mem));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("reset",cmd_reset));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("rst",cmd_reset));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("clear",cmd_clear_port));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("clr",cmd_clear_port));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("c",cmd_clear_port));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("sleep",cmd_sleep));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("info",cmd_info));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("init",cmd_init));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("load",cmd_load));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("save",cmd_save));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("run",cmd_run));
    cmds.insert(std::make_pair<std::string,cmd_func_t>("exit",cmd_exit));

    /* defaults */
    verbose_g = false;
    debug_g   = false;
    forceold_g = false;
    portbase_g = PORT_DEFAULT;
    blocksize_g = 0x0400;
    safe_setup_dtv_g = true;
    file_mode_g = MemBuffer::FILE_MODE_AUTO;

    try {
	/*
	 * First scan for any options.
	 *
	 */
	Getopt::Opt opt;
	Getopt opts(argc, argv, "rlsb:p:fvdVh");
	while (opt = opts.next_opt(), !opt.is_end()) {

	    validate_opt(opt);

	    switch (opt.get_opt()) {

	    /* set raw format */
	    case 'r':
		file_mode_g = MemBuffer::FILE_MODE_RAW;
		break;

	    /* set prg format */
	    case 'l':
		file_mode_g = MemBuffer::FILE_MODE_PRG;
		break;

	    /* disable safe setup dtv */
	    case 's':
		safe_setup_dtv_g = false;
		break;

	    /* set block size */
	    case 'b':
		blocksize_g = get_arg_as_val(opt);
		if (blocksize_g < 1 || blocksize_g > 0x4000) {
		    char tmpstr[256];
		    snprintf(tmpstr, sizeof(tmpstr), "blocksize out of range (=%d, allowable 1-16384)", blocksize_g );
		    throw Exception(tmpstr);
		}
		break;

	    case 'p':
		portbase_g=opt.get_arg();
		break;

	    /* set forceold */
	    case 'f':
		safe_setup_dtv_g = false;
		forceold_g = true;
		break;

	    /* set verbose */
	    case 'v':
		verbose_g = true;
		break;

	    /* set debug */
	    case 'd':
		debug_g = true;
		break;

	    /* print version */
	    case 'V':
		fprintf (stdout, PROGRAM " " VERSION "\n");
		return 0;

	    /* print help */
	    case 'h':
		fprintf (stdout,
PROGRAM " " VERSION "\n"
"Copyright (c) 2005, 2006, 2007, 2008 Daniel Kahlin\n"
"Written by Daniel Kahlin <daniel@kahlin.net>\n"
#ifdef IS_LINUX
"Linux backend written by Hannu Nuotio <hannu.nuotio@pp.inet.fi>\n"
"Copyright (c) 2006, 2007 Hannu Nuotio\n"
#endif
"\n"
"usage: " PROGRAM " [OPTION]... <cmd> [OPTION]... [<file>]\n"
"\n"
"Valid options:\n"
"    -r              use raw format for files (no load address)\n"
"    -l              use prg format for files (has load address)\n"
"    -s              do not enable safe setup timing on the DTV\n"
"    -b <size>       block size (default: 1024)\n"
"    -p <addr>       port address (default: " PORT_DEFAULT ")\n"
"    -f              force pre 1.0 protocol\n"
"    -v              be verbose\n"
"    -d              output debugging information\n"
"    -h              displays this help text\n"
"    -V              output program version\n"
"\n"
"Valid commands:\n"
"    boot/bt/b       send file to boot.prg/mlboot.prg\n"
"    read/rd/r       read mem\n"
"    write/wr/w      write mem\n"
"    verify/vfy      verify mem\n"
"    go/g/sys        execute mem (go)\n"
"    sleep           sleep a number of seconds\n"
"    clear/clr/c     clear port (tristate)\n"
"    info            query server type and configuration\n"
"    init            BASIC init\n"
"    load            BASIC load\n"
"    save            BASIC save\n"
"    run             BASIC run\n"
"    exit            BASIC exit\n"
"  ('dtvtrans <cmd> -h' for more info)\n"
"\n");
		return 0;

	    /* default behavior */
	    default:
		break;
	    }
	}

	opts.set_endmarker('+');
	do {
	    if (opts.args_left() < 1)
		throw Exception("too few arguments");

	    const char* cmd=opts.next_arg();

	    std::map<std::string,cmd_func_t>::const_iterator it=cmds.find(cmd);
	    if (it!=cmds.end()) {
		ret=(it->second)(opts);
	    } else {
		char tmpstr[256];
		snprintf(tmpstr, sizeof(tmpstr), "invalid operation (%s)",cmd);
		throw Exception(tmpstr);
	    }
	} while (!ret && opts.skip_endmarker());

    } catch (std::exception& e) {
	fprintf(stderr, PROGRAM ": %s\n", e.what());
	return 1;
    } catch (...) {
	fprintf(stderr, PROGRAM ": unhandled exception caught!\n");
	return 1;
    }

    return ret;
}

/*************************************************************************
 *
 * NAME  print_xfer_stats
 *
 * DESCRIPTION
 *   print transfer speed and length statistics.
 *
 ******/
void print_xfer_stats(float dur, uint32_t len)
{
    float speed_bytes=(len/dur);
    float speed_bits=(len/dur)*8.0f;
    printf ("transferred %d bytes in %.3f s (%.3f bytes/s, %.3f bit/s)\n",len,dur,speed_bytes,speed_bits);
}

/*************************************************************************
 *
 * NAME  print_buf_stats
 *
 * DESCRIPTION
 *   print buffer statistics.
 *
 ******/
void print_buf_stats(MemBuffer& mb, const char* str)
{
    uint32_t len=mb.get_length();
    int blks=(len+2+253)/254;
    printf("%s... $%06X-$%06X (%d bytes, %d %s)\n",str, mb.get_start(), mb.get_end(),mb.get_length(),blks,(blks==1)?"block":"blocks");
    fflush(stdout);
}

/*************************************************************************
 *
 * NAME  check_protocol_1_0
 *
 * DESCRIPTION
 *   check if protocol revision is at least 1.0 (or disabled by forceold_g)
 *
 ******/
bool check_protocol_1_0(Transfer& tf)
{
    if (forceold_g || tf.get_rev_major() < 0) {
	return false;
    }
    return true;
}

/*************************************************************************
 *
 * NAME  cmd_read_mem
 *
 * DESCRIPTION
 *   Read memory from the server.
 *
 ******/
int cmd_read_mem(Getopt& opts)
{
    Getopt::Opt opt;
    opts.set_optstring("h");
    while (opt = opts.next_opt(), !opt.is_end()) {
	validate_opt(opt);
	switch (opt.get_opt()) {

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... read <range> <file>\n"
"       " PROGRAM " [OPTION]... read <file>\n"
"       " PROGRAM " [OPTION]... read -h\n"
"\n"
"Valid options:\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    // proceed with the command
    if (opts.args_left() < 1)
	throw Exception("too few arguments");

    MemBuffer mb;
    const char* name;
    bool basic_load;

    // if there is exactly one argument, we perform a BASIC save. 
    if (opts.args_left() == 1) {
	name=opts.next_arg();
	basic_load = true;
    } else {
	const char* range=opts.next_arg();
	name=opts.next_arg();
	mb.set_range(range);
	basic_load = false;
    }
	
    Transfer& tf = get_transfer();
    Timer tim;

    if (basic_load) {
	uint32_t sa = tf.read_word(0x002b);
	uint32_t ea = tf.read_word(0x002d);
	mb.set_start(sa);
	mb.set_length(ea-sa);
    }
    print_buf_stats(mb, "reading");
    tim.start();
    tf.read_mem(mb);
    tim.stop();

    print_xfer_stats(tim.get_time_d(), mb.get_length());

    mb.write_file(name, file_mode_g);

    return 0;
}

/*************************************************************************
 *
 * NAME  cmd_verify_mem
 *
 * DESCRIPTION
 *   Read memory from the server and verify against file
 *
 ******/
int cmd_verify_mem(Getopt& opts)
{
    Getopt::Opt opt;
    opts.set_optstring("h");
    while (opt = opts.next_opt(), !opt.is_end()) {
	validate_opt(opt);
	switch (opt.get_opt()) {

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... verify <addr> <file>\n"
"       " PROGRAM " [OPTION]... verify <file>\n"
"       " PROGRAM " [OPTION]... verify -h\n"
"\n"
"Valid options:\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    // proceed with the command
    if (opts.args_left() < 1)
	throw Exception("too few arguments");

    bool basic_verify = false;
    // if there is exactly one argument, we perform a BASIC save. 
    if (opts.args_left() == 1) {
	basic_verify = true;
    }

    MemBuffer mb, refmb;
    if (basic_verify) {
	const char* name=opts.next_arg();  
	refmb.read_file(name, MemBuffer::FILE_MODE_PRG);
    } else {
	const char* addrstr=opts.next_arg();
	const char* name=opts.next_arg();
	refmb.read_file(name, file_mode_g);
	refmb.set_start(addrstr);
    }

    mb.set_start(refmb.get_start());
    mb.set_rom(refmb.is_rom());
    mb.set_length(refmb.get_length());

    Transfer& tf = get_transfer();
    Timer tim;

    print_buf_stats(mb, "verifying");
    tim.start();
    tf.read_mem(mb);
    tim.stop();
    print_xfer_stats(tim.get_time_d(), mb.get_length());

    int len=mb.get_length();
    uint8_t* ptr=mb.get_buffer();
    uint8_t* refptr=refmb.get_buffer();
    if (memcmp(ptr,refptr,len)) {
	printf("data mismatch!\n");
	return 0;
    }
    printf("identical.\n");
    return 0;
}

/*************************************************************************
 *
 * NAME  cmd_write_mem
 *
 * DESCRIPTION
 *   Write file to memory on the server.
 *
 ******/
int cmd_write_mem(Getopt& opts)
{
    Getopt::Opt opt;
    opts.set_optstring("h");
    while (opt = opts.next_opt(), !opt.is_end()) {
	validate_opt(opt);
	switch (opt.get_opt()) {

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... write <addr> <file>\n"
"       " PROGRAM " [OPTION]... write <file>\n"
"       " PROGRAM " [OPTION]... write -h\n"
"\n"
"Valid options:\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    // proceed with the command
    if (opts.args_left() < 1)
	throw Exception("too few arguments");

    MemBuffer mb;
    const char* name;
    // if there is exactly one argument we load as a prg.
    if (opts.args_left() == 1) {
	name=opts.next_arg();  
	mb.read_file(name, MemBuffer::FILE_MODE_PRG);
    } else {
	const char* addrstr=opts.next_arg();
	name=opts.next_arg();  
	mb.read_file(name, file_mode_g);
	mb.set_start(addrstr);
    }

    Transfer& tf = get_transfer();
    Timer tim;

    print_buf_stats(mb, "writing");
    tim.start();
    tf.write_mem(mb);
    tim.stop();

    print_xfer_stats(tim.get_time_d(), mb.get_length());

    return 0;
}

/*************************************************************************
 *
 * NAME  cmd_execute_mem
 *
 * DESCRIPTION
 *   Execute from memory location.
 *
 ******/
int cmd_execute_mem(Getopt& opts)
{
    uint8_t acc=0x00;
    uint8_t xr=0x00;
    uint8_t yr=0x00;
    uint8_t sr=0x00;
    uint8_t mem=0x07;
    uint32_t sa=0xfce2;
    int timeout=-1;

    Getopt::Opt opt;
    opts.set_optstring("a:x:y:s:m:t:h");
    while (opt = opts.next_opt(), !opt.is_end()) {

	validate_opt(opt);

	switch (opt.get_opt()) {

        /* set registers */
        case 'a':
	    acc=get_arg_as_val(opt);
            break;
        case 'x':
	    xr=get_arg_as_val(opt);
            break;
        case 'y':
	    yr=get_arg_as_val(opt);
            break;
        case 's':
	    sr=get_arg_as_val(opt);
            break;
        case 'm':
	    mem=get_arg_as_val(opt);
            break;
        case 't':
	    timeout=get_arg_as_val(opt);
            break;

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... sys [-a<val>][-x<val>][-y<val>][-s<val>][-m<val>][-t<time>] <addr>\n"
"       " PROGRAM " [OPTION]... sys -h\n"
"\n"
"Valid options:\n"
"    -a              set accumulator (default: 0x00)\n"
"    -x              set X register (default: 0x00)\n"
"    -y              set Y register (default: 0x00)\n"
"    -s              set status reg  (default: 0x00)\n"
"    -m              set memory ($01) (default: 0x07)\n"
"    -t              set timeout for getting result (default: -1)\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    if (opts.args_left() < 1)
	throw Exception("too few arguments");

    const char* start=opts.next_arg();
    char* endptr;

    sa = std::strtoul(start,&endptr,0);

    Transfer& tf = get_transfer();

    printf("executing from $%04X (%d)\n",sa ,sa);
    if (!check_protocol_1_0(tf)) {
	tf.execute_mem(sa);
	return 0;
    }

    tf.sys(sa, acc, xr, yr, sr, mem);

    if (timeout > 0) {
	if (!tf.check_present(timeout)) {
	    printf("didn't return!\n");
	    return 0;
	}
	tf.sys_result(acc, xr, yr, sr);
	printf("result: Acc=$%02X, X=$%02X, Y=$%02X, SR=$%02X\n", acc, xr, yr, sr);
    }
    return 0;
}

/*************************************************************************
 *
 * NAME  cmd_boot
 *
 * DESCRIPTION
 *   Send file to server using the boot protocol.
 *
 ******/
int cmd_boot(Getopt& opts)
{
    Getopt::Opt opt;
    opts.set_optstring("h");
    while (opt = opts.next_opt(), !opt.is_end()) {
	validate_opt(opt);
	switch (opt.get_opt()) {

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... boot <file>\n"
"       " PROGRAM " [OPTION]... boot -h\n"
"\n"
"Valid options:\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    // proceed with the command
    if (opts.args_left() < 1)
	throw Exception("too few arguments");

    const char* name=opts.next_arg();

    MemBuffer mb;
    mb.read_file(name, MemBuffer::FILE_MODE_PRG);

    uint32_t sa=mb.get_start();
    uint32_t len=mb.get_length();

    Transfer& tf = get_transfer();
    Timer tim;

    printf("sending boot... ($%04X-$%04X)\n",sa,sa+len-1);
    fflush(stdout);
    tim.start();
    tf.send_file_boot(sa,sa+len-1, mb.get_buffer());
    tim.stop();
   
    print_xfer_stats(tim.get_time_d(), len);

    return 0;
}

/*************************************************************************
 *
 * NAME  cmd_reset
 *
 * DESCRIPTION
 *   Try to reset into server mode.
 *
 ******/
int cmd_reset(Getopt& opts)
{
    bool knock = true;
    bool do_init = true;
    Transfer::init_mode_t mode=Transfer::INIT_FULL_BASIC;
    Getopt::Opt opt;
    opts.set_optstring("NnBbKkh");
    while (opt = opts.next_opt(), !opt.is_end()) {

	validate_opt(opt);

	switch (opt.get_opt()) {

        /* set reset modes */
        case 'N':
	    knock = false;
            break;
        case 'n':
	    do_init = false;
            break;

        /* set init modes */
        case 'B':
	    mode = Transfer::INIT_FULL_BASIC;
            break;
        case 'b':
	    mode = Transfer::INIT_MINIMAL_BASIC;
            break;
        case 'K':
	    mode = Transfer::INIT_FULL_KERNAL;
            break;
        case 'k':
	    mode = Transfer::INIT_MINIMAL_KERNAL;
            break;

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... reset [-N][-n][-B][-b][-K][-k]\n"
"       " PROGRAM " [OPTION]... reset -h\n"
"\n"
"Valid options:\n"
"    -N              Reset only, don't enter dtvtrans\n"
"    -n              Reset into dtvtrans, no init\n"
"    -B              Full BASIC (default)\n"
"    -b              Minimal BASIC\n"
"    -K              Full KERNAL\n"
"    -k              Minimal KERNAL\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    Transfer& tf = get_transfer();
    if (knock)
	printf("resetting into dtvtrans\n");
    else
	printf("resetting\n");
    tf.reset(knock);
    if (knock) {
	if (!tf.check_present(1)) {  
	    return 0;
	}
    }


    if (knock && do_init) {
	if (check_protocol_1_0(tf)) {
	    perform_init(mode);
	}
    }

    return 0;
}

/*************************************************************************
 *
 * NAME  cmd_clear_port
 *
 * DESCRIPTION
 *   Clear/tri-state port bits.
 *
 ******/
int cmd_clear_port(Getopt& opts)
{
    Getopt::Opt opt;
    opts.set_optstring("h");
    while (opt = opts.next_opt(), !opt.is_end()) {
	validate_opt(opt);
	switch (opt.get_opt()) {

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... clear\n"
"       " PROGRAM " [OPTION]... clear -h\n"
"\n"
"Valid options:\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    // proceed with the command
    Transfer& tf = get_transfer();
    printf("tri-stating port\n");
    tf.tristate();

    return 0;
}

/*************************************************************************
 *
 * NAME  cmd_sleep
 *
 * DESCRIPTION
 *   Sleep for <arg> seconds.   fractions accepted.
 *
 ******/
int cmd_sleep(Getopt& opts)
{
    Getopt::Opt opt;
    opts.set_optstring("h");
    while (opt = opts.next_opt(), !opt.is_end()) {
	validate_opt(opt);
	switch (opt.get_opt()) {

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... sleep <time>\n"
"       " PROGRAM " [OPTION]... sleep -h\n"
"\n"
"Valid options:\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    // proceed with the command
    if (opts.args_left() < 1)
	throw Exception("too few arguments");

    float t=std::atof(opts.next_arg());

    printf("sleeping for %.1f %s...\n", t, t==1.0?"second":"seconds");
    fflush(stdout);
    SSleep slp;
    slp.msleep(static_cast<int>(t*1000));

    return 0;
}

/*************************************************************************
 *
 * NAME  cmd_info
 *
 * DESCRIPTION
 *   Determine server implementation and revision.
 *
 ******/
int cmd_info(Getopt& opts)
{
    Getopt::Opt opt;
    opts.set_optstring("h");
    while (opt = opts.next_opt(), !opt.is_end()) {
	validate_opt(opt);
	switch (opt.get_opt()) {

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... info\n"
"       " PROGRAM " [OPTION]... info -h\n"
"\n"
"Valid options:\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    // proceed with the command
    Transfer& tf = get_transfer();

    if (!tf.check_present(1))
	throw Exception("server not present!");
    printf("server presence detected.\n");

    printf("implementation: %s\n", tf.get_implementation());
    int maj=tf.get_rev_major();
    int min=tf.get_rev_minor();
    if (maj < 0) {
	printf("protocol revision: pre 1.0\n");
	return 0;
    }

    printf("protocol revision: %d.%d\n", maj, min);

    const char* str="<unknown>";
    switch (tf.get_conf_port()) {
    case 0:
	str="Joystick 1";
	break;
    case 1:
	str="Joystick 2";
	break;
    case 2:
	str="Userport";
	break;
    default:
	break;
    }
    printf("port: %s\n", str);
    printf(
        "location: $%06X-$%06X (in %s)\n",
	tf.get_conf_start(), tf.get_conf_end(),
	tf.get_conf_rom() ? "Flash":"RAM"
    );
    return 0;
}

/*************************************************************************
 *
 * NAME  cmd_init, perform_init
 *
 * DESCRIPTION
 *   Initialize DTV environment.
 *
 ******/
int cmd_init(Getopt& opts)
{
    Transfer::init_mode_t mode=Transfer::INIT_FULL_BASIC;
    Getopt::Opt opt;
    opts.set_optstring("BbKkh");
    while (opt = opts.next_opt(), !opt.is_end()) {

	validate_opt(opt);

	switch (opt.get_opt()) {

        /* set init modes */
        case 'B':
	    mode = Transfer::INIT_FULL_BASIC;
            break;
        case 'b':
	    mode = Transfer::INIT_MINIMAL_BASIC;
            break;
        case 'K':
	    mode = Transfer::INIT_FULL_KERNAL;
            break;
        case 'k':
	    mode = Transfer::INIT_MINIMAL_KERNAL;
            break;

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... init [-B][-b][-K][-k]\n"
"       " PROGRAM " [OPTION]... init -h\n"
"\n"
"Valid options:\n"
"    -B              Full BASIC (default)\n"
"    -b              Minimal BASIC\n"
"    -K              Full KERNAL\n"
"    -k              Minimal KERNAL\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    Transfer& tf = get_transfer();

    if (!check_protocol_1_0(tf))
	throw Exception("not supported by server (init)");

    perform_init(mode);
    return 0;
}

void perform_init(Transfer::init_mode_t mode)
{
    Transfer& tf = get_transfer();

    if (debug_g)
	printf("init(mode=$%02X)\n",mode);

    if (mode == Transfer::INIT_FULL_BASIC || mode == Transfer::INIT_MINIMAL_BASIC) {
	printf("initializing BASIC...\n");
    } else {
	printf("initializing KERNAL...\n");
    }
    tf.init(mode);
    if (mode == Transfer::INIT_FULL_BASIC || mode == Transfer::INIT_MINIMAL_BASIC) {
	tf.print(
	    CBM_CR
	    "    **** COMMODORE 64 BASIC V2 ****" CBM_CR
	    CBM_CR
	    " 64K RAM SYSTEM  38911 BASIC BYTES FREE" CBM_CR
	    CBM_READY
	    CBM_CURSOR
	);
    }
}

/*************************************************************************
 *
 * NAME  cmd_load
 *
 * DESCRIPTION
 *   Perform "load".  If no arguments, simulate load of the data between
 *   $2b and $2d in ram.
 *   Otherwise transfer the file into ram, set up $ae/$2d and simulate load.
 *
 ******/
int cmd_load(Getopt& opts)
{
    bool simulate_load = true;
    Transfer::load_mode_t load_mode = Transfer::LOAD_NORMAL;

    Getopt::Opt opt;
    opts.set_optstring("Nnh");
    while (opt = opts.next_opt(), !opt.is_end()) {
	validate_opt(opt);
	switch (opt.get_opt()) {

        /* set load modes */
        case 'N':
	    simulate_load = false;
            break;
        case 'n':
	    load_mode = Transfer::LOAD_NO_RELINK;
            break;

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... load [-N][-n][<file>]\n"
"       " PROGRAM " [OPTION]... load -h\n"
"\n"
"Valid options:\n"
"    -N              do not simulate load, only set end address\n"
"    -n              do not perform relink, only clr\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    // proceed with the command
    uint32_t ea;
    Transfer& tf = get_transfer();

    if (opts.args_left() > 0) {
	const char* name;
	MemBuffer mb;
	name=opts.next_arg();  
	mb.read_file(name, MemBuffer::FILE_MODE_PRG);

	Timer tim;

	print_buf_stats(mb, "loading");
	tim.start();
	tf.write_mem(mb);

	ea=mb.get_end();
	tf.write_word(0x00ae, ea);
	tf.write_word(0x002d, ea);
	tf.write_byte(0x00ba, 8);  //device number
	tim.stop();

	print_xfer_stats(tim.get_time_d(), mb.get_length());

    } else {
	ea=tf.read_word(0x002d);
    }

    if (simulate_load && check_protocol_1_0(tf)) {
	printf("simulating load $0801-$%04X...\n",ea);
	tf.print(
	    CBM_NOCURSOR
	    "LOAD\"REMOTE\",8,1" CBM_CR CBM_CR
	    "SEARCHING FOR REMOTE" CBM_CR
	    "LOADING"
	    CBM_READY
	    CBM_CURSOR
	);
	tf.load(load_mode);
    }
    return 0;
}

/*************************************************************************
 *
 * NAME  cmd_save
 *
 * DESCRIPTION
 *   Perform "save" by fetching the addresses in $2b and $2d, and then
 *   reading the data between those to a file.
 *
 ******/
int cmd_save(Getopt& opts)
{
    Getopt::Opt opt;
    opts.set_optstring("h");
    while (opt = opts.next_opt(), !opt.is_end()) {
	validate_opt(opt);
	switch (opt.get_opt()) {

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... save <file>\n"
"       " PROGRAM " [OPTION]... save -h\n"
"\n"
"Valid options:\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    // proceed with the command
    if (opts.args_left() < 1)
	throw Exception("too few arguments");

    const char* name=opts.next_arg();

    Transfer& tf = get_transfer();
    MemBuffer mb;
    uint32_t sa,ea;

    sa = tf.read_word(0x002b);
    ea = tf.read_word(0x002d);
    mb.set_start(sa);
    mb.set_length(ea-sa);

    Timer tim;

    print_buf_stats(mb, "saving");
    tim.start();
    tf.read_mem(mb);
    tim.stop();

    print_xfer_stats(tim.get_time_d(), mb.get_length());

    mb.write_file(name, MemBuffer::FILE_MODE_PRG);

    if (check_protocol_1_0(tf)) {
	tf.print(
	    CBM_NOCURSOR
	    "SAVE\"REMOTE\",8" CBM_CR CBM_CR
	    "SAVING REMOTE"
	    CBM_READY
	    CBM_CURSOR
	);
    }
    return 0;
}

/*************************************************************************
 *
 * NAME  cmd_run
 *
 * DESCRIPTION
 *   Make server perform BASIC RUN.
 *
 ******/
int cmd_run(Getopt& opts)
{
    Getopt::Opt opt;
    opts.set_optstring("h");
    while (opt = opts.next_opt(), !opt.is_end()) {
	validate_opt(opt);
	switch (opt.get_opt()) {

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... run\n"
"       " PROGRAM " [OPTION]... run -h\n"
"\n"
"Valid options:\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    // proceed with the command
    Transfer& tf = get_transfer();

    if (!check_protocol_1_0(tf))
	throw Exception("not supported by server (run)");

    printf("performing RUN...\n");
    tf.print(
	CBM_NOCURSOR
	"RUN" CBM_CR
    );
    tf.run(Transfer::RUN_NORMAL);
    return 0;
}

/*************************************************************************
 *
 * NAME  cmd_exit
 *
 * DESCRIPTION
 *   Make server exit to BASIC.
 *
 ******/
int cmd_exit(Getopt& opts)
{
    Transfer::exit_mode_t exit_mode = Transfer::EXIT_NORMAL;

    Getopt::Opt opt;
    opts.set_optstring("nh");
    while (opt = opts.next_opt(), !opt.is_end()) {
	validate_opt(opt);
	switch (opt.get_opt()) {

	/* set exit modes */
        case 'n':
	    exit_mode = Transfer::EXIT_ONLY_EXIT;
            break;

        /* print help */
        case 'h':
            fprintf (stdout,
"usage: " PROGRAM " [OPTION]... exit [-n]\n"
"       " PROGRAM " [OPTION]... exit -h\n"
"\n"
"Valid options:\n"
"    -n              only exit, do not perform clr\n"
"    -h              displays this help text\n"
"\n");
            return 0;

        /* default behavior */
        default:
            break;
        }
    }

    // proceed with the command
    Transfer& tf = get_transfer();

    if (!check_protocol_1_0(tf))
	throw Exception("not supported by server (exit)");

    printf("exiting to BASIC...\n");
    tf.print(
	CBM_NOCURSOR
    );
    tf.exit(exit_mode);
    return 0;
}
/* eof */
