/*************************************************************************
 *
 * FILE  parport_transfer.cpp
 * Copyright (c) 2006, 2008 Daniel Kahlin <daniel@kahlin.net>
 * Written by Daniel Kahlin <daniel@kahlin.net>
 * $Id: parport_transfer.cpp,v 1.34 2008-02-17 22:45:36 tlr Exp $
 *
 * DESCRIPTION 
 *
 ******/

#include <algorithm>
#include <stdio.h>
#include <cstring>
#include "global.hpp"

#include "parport_transfer.hpp"
// arch dependant
#include "timer.hpp"

static USleep sleeper;

ParportTransfer::ParportTransfer(const std::string& portbase, int blocksize, bool use_safe_setup_dtv)
    : Transfer(blocksize, use_safe_setup_dtv), pt_(portbase)
{
}

void ParportTransfer::read_mem_raw(uint8_t mode, uint8_t bank, uint16_t base, uint16_t len, uint8_t* target)
{
    // command read_mem
    send_byte_(CMD_READ_MEM);

    // mode (0=ram, 1=rom)
    send_byte_(mode);

    // bank
    send_byte_(bank);

    // addr
    send_byte_(base & 0xff);
    send_byte_(base >> 8);

    // len 
    send_byte_(len & 0xff);
    send_byte_(len >> 8);


    uint8_t chk=0;
    for (int i=0; i < len; ++i) {
	uint8_t tmp=recv_byte_();
	chk+=tmp+1;
	target[i]=tmp;
    }
    uint8_t chkb=recv_byte_();

    if (chk != chkb)
	printf("ERROR: got block, chk=$%02X (expected $%02X)\n",chk,chkb);
}

void ParportTransfer::write_mem_raw(uint8_t mode, uint8_t bank, uint16_t base, uint16_t len, uint8_t* source)
{
    // command write_mem
    send_byte_(CMD_WRITE_MEM);

    // mode (0=ram, 1=rom (invalid))
    send_byte_(mode);

    // bank
    send_byte_(bank);

    // addr
    send_byte_(base & 0xff);
    send_byte_(base >> 8);

    // len 
    send_byte_(len & 0xff);
    send_byte_(len >> 8);

    uint8_t chk=0;
    for (int i=0; i < len; ++i) {
	uint8_t tmp=source[i];
	send_byte_(tmp);
	chk+=tmp+1;
    }
    send_byte_(chk);
    uint8_t chkb=recv_byte_();

    if (chk != chkb)
	printf("ERROR: sent block, chk=$%02X (receiver got $%02X)\n",chk,chkb);
}

void ParportTransfer::execute_mem(uint16_t addr)
{
    get_revision_();

    // command execute_mem
    send_byte_(CMD_EXECUTE_MEM);
    // addr
    send_byte_(addr & 0xff);
    send_byte_(addr >> 8);

    release_();
    // here the timing is reset to the defaults on the DTV side as
    // the server does a reinit upon returning from the code.
    clear_state_();
}

void ParportTransfer::sys(uint16_t addr, uint8_t acc, uint8_t xr, uint8_t yr, uint8_t sr, uint8_t mem)
{
    send_byte_(CMD_SYS);
    // mode (00=normal)
    send_byte_(0x00);
    // addr
    send_byte_(addr & 0xff);
    send_byte_(addr >> 8);
    // regs
    send_byte_(sr);
    send_byte_(acc);
    send_byte_(xr);
    send_byte_(yr);
    send_byte_(mem);

    release_();
    // here the timing is reset to the defaults on the DTV side as
    // the server does a reinit upon returning from the code.
    clear_state_();
}

void ParportTransfer::sys_result(uint8_t& acc, uint8_t& xr, uint8_t& yr, uint8_t& sr)
{
    send_byte_(CMD_SYS_RESULT);
    // mode (00=normal)
    send_byte_(0x00);
    // get parameters
    sr = recv_byte_();
    acc = recv_byte_();
    xr = recv_byte_();
    yr = recv_byte_();
}


void ParportTransfer::init(init_mode_t mode)
{
    // command init
    send_byte_(CMD_INIT);
    // mode
    send_byte_(mode);

    release_();
}

void ParportTransfer::load(load_mode_t mode)
{
    // command load
    send_byte_(CMD_LOAD);
    // mode
    send_byte_(mode);

    release_();
}

void ParportTransfer::run(run_mode_t mode)
{
    // command run
    send_byte_(CMD_RUN);
    // mode
    send_byte_(mode);

    release_();
    // here the timing is reset to the defaults on the DTV side as
    // the server does a reinit upon returning from the code.
    clear_state_();
}

void ParportTransfer::exit(exit_mode_t mode)
{
    // command exit
    send_byte_(CMD_EXIT);
    // mode
    send_byte_(mode);

    release_();
    // here the timing is reset to the defaults on the DTV side as
    // the server does a reinit upon returning from the code.
    clear_state_();
}

void ParportTransfer::print_raw(const char* str, int len)
{
    // command exit
    send_byte_(CMD_PRINT);
    // len
    send_byte_(len);
    for (int i=0; i < len; ++i)
	send_byte_(str[i]);
}

void ParportTransfer::color(uint8_t border, uint8_t background, uint8_t text)
{
    // command exit
    send_byte_(CMD_COLOR);
    send_byte_(border);
    send_byte_(background);
    send_byte_(text);
}

#define PRESENT_STEP 2
#define IDLE_STEP 20
bool ParportTransfer::check_present(int timeout)
{
    int t=timeout*1000;
    SSleep slp;
    int state=0;

    while (t > 0) {
	switch (state) {
	case 0:
	    // clk = 0
	    pt_.write(0xf7);
	    if ((pt_.read() & 0x10) == 0) {
		slp.msleep(PRESENT_STEP); t-=PRESENT_STEP;
		if ((pt_.read() & 0x10) == 0) {
		    slp.msleep(PRESENT_STEP); t-=PRESENT_STEP;
		    if ((pt_.read() & 0x10) == 0) {
			state=1;
		    }
		}
	    }
	    break;
	case 1:
	    // clk = 1
	    pt_.write(0xff);
	    if ((pt_.read() & 0x10) == 0x10) {
		slp.msleep(PRESENT_STEP); t-=PRESENT_STEP;
		if ((pt_.read() & 0x10) == 0x10) {
		    slp.msleep(PRESENT_STEP); t-=PRESENT_STEP;
		    if ((pt_.read() & 0x10) == 0x10) {
			state=2;
		    }
		}
	    }
	    break;
	case 2:
	    // clk = 0
	    pt_.write(0xf7);
	    if ((pt_.read() & 0x10) == 0) {
		slp.msleep(PRESENT_STEP); t-=PRESENT_STEP;
		if ((pt_.read() & 0x10) == 0) {
		    slp.msleep(PRESENT_STEP); t-=PRESENT_STEP;
		    if ((pt_.read() & 0x10) == 0) {
			state=3;
		    }
		}
	    }
	    break;
	case 3:
	    // clk = 1
	    pt_.write(0xff);
	    if ((pt_.read() & 0x10) == 0x10) {
		slp.msleep(PRESENT_STEP); t-=PRESENT_STEP;
		if ((pt_.read() & 0x10) == 0x10) {
		    slp.msleep(PRESENT_STEP); t-=PRESENT_STEP;
		    if ((pt_.read() & 0x10) == 0x10) {
			return true;
		    }
		}
	    }
	    break;
	default:
	    break;
	}
	slp.msleep(IDLE_STEP); t-=IDLE_STEP;
    }
    return false;
}

void ParportTransfer::tristate()
{
    pt_.tristate();
}

void ParportTransfer::reset(bool knock)
{
    SSleep slp;

    if (knock) {
	// rst=0
	pt_.write(0xdf);
	slp.msleep(10);
	// rst=0, ack=0, d0=0
	pt_.write(0xce);
	slp.msleep(10);
	// rst=1, ack=0, d0=0
	pt_.write(0xee);
	slp.msleep(500);
	// rst=1, ack=1, d0=1
	pt_.write(0xff);
    } else {
	// rst=0
	pt_.write(0xdf);
	slp.msleep(20);
	// rst=1
	pt_.write(0xff);
    }

    // restore states that are now undetermined
    clear_state_();
}

const char* ParportTransfer::get_implementation()
{
    get_revision_();

    if (prev_major_ < 0) {
	return "<unknown>";
    }
    // command query implementation
    send_byte_(CMD_QUERY_IMPLEMENTATION);

    uint8_t len = recv_byte_();

    int i;
    for (i=0; i<len; i++) {
	impl_str_[i]=recv_byte_();
    }
    impl_str_[i]=0;

    return impl_str_;
}

void ParportTransfer::get_revision_raw()
{
    // command query revision
    send_byte_(CMD_QUERY_REVISION);
    // receive the first byte 
    // This will be it will be 0xff for earlier versions because the
    // server will not accept the above command, waiting for a new command.
    // A receive byte here will get data from the tri-stated server,
    // resulting in all 1's.
    uint8_t major = recv_byte_();
    if (major == 0xff) {
	prev_major_ = -1;
	prev_minor_ = -1;
    } else {
	prev_major_ = major;
	prev_minor_ = recv_byte_();
    }
}

void ParportTransfer::get_configuration_raw()
{
    // command 
    send_byte_(CMD_QUERY_CONFIG);

    conf_port_ = recv_byte_();

    conf_rom_ = recv_byte_();
    conf_sa_ = recv_byte_();
    conf_sa_ |= recv_byte_() << 8;
    conf_sa_ |= recv_byte_() << 16;
    conf_ea_ = recv_byte_();
    conf_ea_ |= recv_byte_() << 8;
    conf_ea_ |= recv_byte_() << 16;
}

void ParportTransfer::set_timing_raw(bool safe_setup_dtv)
{
    send_byte_(CMD_SET_BLOCKCTRL);
    send_byte_( (safe_setup_dtv)?0x01:0x00 );
}


void ParportTransfer::release_()
{
    // tri-state
    pt_.tristate();
    // allow uninit to detect the port as free, and allow for some
    // initial toggling on the DTV side.
    SSleep slp;
    slp.msleep(1);
}
 
void ParportTransfer::send_byte_(uint8_t data)
{
    uint8_t tmp;
    // bit7-5 to d2/d1/d0, clk=0
    tmp=(data >> 5) & 0x07;
    tmp|=0x10;  // keep ACK high to be able to read it.
    tmp|=0x20;  // rst=1.
    pt_.write(tmp);
    while( pt_.read()&0x10 ); // wait for ACK low

    // bit4-2 to d2/d1/d0, clk=1
    tmp=(data >> 2) & 0x07;
    tmp|=0x08;
    tmp|=0x10;  // keep ACK high to be able to read it.
    tmp|=0x20;  // rst=1.
    pt_.write(tmp);
    while( !(pt_.read()&0x10) );  // wait for ACK high

    // bit1-0 to d1/d0  d2=0, clk=0
    tmp=(data >> 0) & 0x03;
    tmp|=0x10;  // keep ACK high to be able to read it.
    tmp|=0x20;  // rst=1.
    pt_.write(tmp);
    while( pt_.read()&0x10 );  // wait for ACK low

    // d2/d1/d0=1, clk=1
    tmp=0x0f;
    tmp|=0x10;  // keep ACK high to be able to read it.
    tmp|=0x20;  // rst=1.
    pt_.write(tmp);
    while( !(pt_.read()&0x10) );  // wait for ACK high

    // tri-state
    pt_.tristate();
    // done!
}

uint8_t ParportTransfer::recv_byte_()
{
    uint8_t data;
    uint8_t tmp;

    // ack/d2/d1/d0=1, clk=0, rst=1
    pt_.write(0x37);
    while( tmp = pt_.read(), tmp&0x10 ); // wait for ACK low
    if (safe_timing_pc_) {
	sleeper.usleep(5);        // this is to allow for setup on the signals
	tmp=pt_.read();
    }
    data = (tmp & 0x07) << 5;

    // ack/d2/d1/d0=1, clk=1, rst=1
    pt_.write(0x3f);
    while( tmp = pt_.read(), !(tmp&0x10) );  // wait for ACK high
    if (safe_timing_pc_) {
	sleeper.usleep(5);        // this is to allow for setup on the signals
	tmp=pt_.read();
    }
    data |= (tmp & 0x07) << 2;

    // ack/d2/d1/d0=1, clk=0, rst=1
    pt_.write(0x37);
    while( tmp = pt_.read(), tmp&0x10 );  // wait for ACK low
    if (safe_timing_pc_) {
	sleeper.usleep(5);        // this is to allow for setup on the signals
	tmp=pt_.read();
    }
    data |= (tmp & 0x03) << 0;

    // ack/d2/d1/d0=1, clk=1, rst=1
    pt_.write(0x3f);
    while( !(pt_.read()&0x10) );  // wait for ACK high

    // done!
    return data;
}



void ParportTransfer::send_file_boot(uint16_t start, uint16_t end, uint8_t* data)
{
    send_byte_boot_(start >> 8);
    send_byte_boot_(start & 0xff);
    send_byte_boot_(end >> 8);
    send_byte_boot_(end & 0xff);

    uint8_t ck=0;
    for (int i=start; i<=end; ++i) {
	send_byte_boot_(*data);
	ck+=(*data)+1;
	++data;
    }
    send_byte_boot_(ck);
    pt_.tristate();
}

void ParportTransfer::send_byte_boot_(uint8_t data)
{
    uint8_t tmp;
    // bit7-6 to d1/d0, clk=0
    tmp=(data >> 6) & 0x03;
    tmp|=0x10;  // keep ACK high to be able to read it.
    tmp|=0x20;  // rst=1.
    pt_.write(tmp);
    while( pt_.read()&0x10 ); // wait for ACK low

    // bit5-4 to d1/d0, clk=1
    tmp=(data >> 4) & 0x03;
    tmp|=0x08;
    tmp|=0x10;  // keep ACK high to be able to read it.
    tmp|=0x20;  // rst=1.
    pt_.write(tmp);
    while( !(pt_.read()&0x10) );  // wait for ACK high

    // bit3-2 to d1/d0, clk=0
    tmp=(data >> 2) & 0x03;
    tmp|=0x10;  // keep ACK high to be able to read it.
    tmp|=0x20;  // rst=1.
    pt_.write(tmp);
    while( pt_.read()&0x10 );  // wait for ACK low

    // bit1-0 to d1/d0, clk=1
    tmp=(data >> 0) & 0x03;
    tmp|=0x08;
    tmp|=0x10;  // keep ACK high to be able to read it.
    tmp|=0x20;  // rst=1.
    pt_.write(tmp);
    while( !(pt_.read()&0x10) );  // wait for ACK high
}

/* eof */
